
# ARIMA Cross-validation by rolling windows
# Adapted from Rob J Hyndman's code:
# http://robjhyndman.com/hyndsight/tscvexample/
#
# Could use further generalization, e.g. to seasonality
# Careful!  This can produce singularities using categorical covariates
arimaCV <- function(x, order, seasonal, xreg, include.mean, forward=1, minper=50) {
  require(forecast)
  if (!any(class(x)=="ts")) x <- ts(x)
  n <- length(x)
  mae <- matrix(NA, nrow=n-minper, ncol=forward)
  st <- tsp(x)[1]+(minper-2)   
  for(i in 1:(n-minper)) {
    xshort <- window(x, start=st+(i-minper+1), end=st+i)
    xnext <- window(x, start=st+(i+1), end=min(n, st+(i+forward)))
    xregshort <- window(xreg, start=st+(i-minper+1), end=st+i)
    xregnext <- window(xreg, start=st+(i+1), end=min(n, st+(i+forward)))
    fit <- Arima(xshort, 
                 order=order, 
                 seasonal=seasonal, 
                 xreg=xregshort, 
                 include.mean=include.mean)
    fcast <- forecast(fit, h=length(xnext), xreg=xregnext)
    mae[i,1:length(xnext)] <- abs(fcast[['mean']]-xnext)
  }
  colMeans(mae, na.rm=TRUE)
}


# Simple univariate time series simulator:
# Allows deterministic trends, seasonality, and ARMA(p,q)
# process, including potentially nonstationary and/or
# noninvertible processes, as well as covariates
#
# Christopher Adolph (faculty.washington.edu/cadolph)
# 30 May 2015
#
# Usage:
#
# y <- makeTS(n = 1000,      # Simulate 1000 periods
#             ar = c(0.67),  # AR(1); phi = 0.67
#             ma = NULL,     # MA(0)
#             trend = 0.01,  # Determ. trend (+0.01 units/period)
#             seasons = c(0,0,0, 0,0,0, 0,0,0, 0,0,0), # 12 period cycle
#             adjust = c("level"), # Additive cycles; alternative is "factor"
#                                  # which multiplies last round
#             x = NULL,      # n x k matrix of exogenous covariates
#             beta = NULL,   # k-vector of parameters relating x to y
#             varY = 1,      # Error term variance
#             initY = 0,     # Prior level of time series, for lags
#             initE = 0,     # Prior error, for lags
#             burnin = 10    # Discard this many iterations before saving
#             )
#



makeTS <- function(n=1,
                   ar=NULL,
                   ma=NULL,
                   trend=NULL,
                   seasons=NULL,
                   adjust=c("factor", "level"),
                   x=NULL,
                   beta=NULL,
                   varY=1,
                   initY=0,
                   initE=0,
                   burnin=0
) {
  
  nseasons <- length(seasons)
  doS <- length(seasons)>0
  doT <- length(trend)>0
  doA <- length(ar)>0
  doM <- length(ma)>0
  
  if (!is.null(x)) {doX <- TRUE; x <- as.matrix(x)} else {doX <- FALSE}
  
  y <- rep(NA, n + burnin)
  
  while (length(initY)<length(ar)) initY <- c(initY,initY)
  lagY <- initY[1:length(ar)]
  
  while (length(initE)<length(ma)) initE <- c(initE,initE)
  lagE <- initE[1:length(ma)]
  
  Scomp <- Tcomp <- Acomp <- Mcomp <- Xcomp <- 0
  for (i in 1:(n+burnin)) {
    
    if (doT) Tcomp <- trend[1]*i
    
    if (doA) Acomp <- sum(ar*rev(lagY))
    
    if (doM) Mcomp <- sum(ma*rev(lagE))
    
    if (doX) Xcomp <- x[i,]%*%beta
    
    E <- rnorm(1)*sqrt(varY)
    
    y[i] <- Tcomp + Acomp + Mcomp + Xcomp + E
    
    if (doS) {
      currSeas <- seasons[((i%%nseasons==0)*nseasons + (i%%nseasons))]
      if (adjust[1]=="factor") {
        y[i] <- y[i]*currSeas
      } else {
        y[i] <- y[i] + currSeas
      }
    } 
    
    lagY <- c(lagY, y[i])
    lagY <- lagY[2:length(lagY)]
    
    lagE <- c(lagE, E)
    lagE <- lagE[2:length(lagE)]
    
  }
  
  y[(burnin+1):(n+burnin)]
}




# Simple time series plots of data and model-based forecasts
# (uses R base graphics)
#
# Allows for up to 2 forecasts, with CIs or 2*se intervals
# and allows plotting of "factual" data over the forecast period
#
# Christopher Adolph
# 28 July 2014
#
ctrfactTS <- function(observed,
                      predicted,
                      lower=NULL,
                      upper=NULL,
                      se=NULL,
                      predicted0=NULL,
                      lower0=NULL,
                      upper0=NULL,
                      se0=NULL,
                      factual=NULL,
                      at.xaxis=NULL,
                      lab.xaxis=NULL,
                      ylim=NULL,
                      main="Counterfactual time series",
                      xlab="x",
                      ylab="y",
                      col="red",
                      col0="blue"
) {
  
  
  ndata <- length(observed)
  ncft <- max(c(length(predicted), length(predicted0), length(factual)))
  
  if (!is.null(se)) {
    lower <- predicted - 2*se
    upper <- predicted + 2*se
  }
  if (!is.null(se0)) {
    lower0 <- predicted0 - 2*se0
    upper0 <- predicted0 + 2*se0
  }
  
  if (is.null(ylim)) {
    ydata <- c(observed, predicted, lower, upper, predicted0, lower0, upper0)
    ylim <- c(min(ydata), max(ydata))
  }
  
  plot.new()
  par(usr = c(0, ndata+ncft, ylim[1], ylim[2]) )
  
  # make the x-axis
  if (!is.null(at.xaxis)) {
    if (!is.null(lab.xaxis)) {
      axis(1, at=at.xaxis, labels=lab.xaxis)
    } else {
      axis(1, at=at.xaxis)
    }
  } else {
    axis(1)
  }
  
  # Make the y-axis
  axis(2)
  
  # Make titles
  title(main=main, xlab=xlab, ylab=ylab)
  
  # Polygon of confidence interval
  if (!is.null(lower)) {
    x0 <- (ndata+1):(ndata + length(lower))
    polygon(x = c(ndata, x0, rev(x0), ndata),
            y = c(observed[ndata], lower, rev(upper), observed[ndata]),
            border=NA,
            col=lighten(col)
    )
  }
  
  # Plot the actual data (including factual data if present)
  if (!is.null(factual)) {
    lines(x = 1:(ndata+length(factual)),
          y = c(observed, factual))
  } else {
    lines(x = 1:ndata,
          y = observed
    )
  }
  
  # Add the predictions
  lines(x = ndata:(ndata+length(predicted)),
        y = c(observed[length(observed)],predicted),  # link up the actual data to the prediction
        col = col
  )
  
  # Add the lower predictive interval for alt scenario
  if (!is.null(lower0)) {
    lines(x = ndata:(ndata+length(predicted0)),
          y = c(observed[length(observed)],lower0),
          col = col0,
          lty = "dashed"
    )
  }
  
  # Add the upper predictive interval for no law
  if (!is.null(upper0)) {
    lines(x = ndata:(ndata+length(predicted0)),
          y = c(observed[length(observed)],upper0),
          col = col0,
          lty = "dashed"
    )
  }
  
  # Add the predictions for keeping law
  if (!is.null(predicted0)) {
    lines(x = ndata:(ndata+length(predicted0)),
          y = c(observed[length(observed)],predicted0), 
          col = col0
    )
  }
  
}




lighten <- function (col, pct = 0.75, alpha = 1) 
{
  if (abs(pct) > 1) {
    print("Warning:  Error in Lighten; invalid pct")
    pcol <- col2rgb(col)/255
  }
  else {
    col <- col2rgb(col)/255
    if (pct > 0) {
      pcol <- col + pct * (1 - col)
    }
    else {
      pcol <- col * pct
    }
  }
  pcol <- rgb(pcol[1], pcol[2], pcol[3], alpha)
  pcol
}
